import importlib
import inspect
from pathlib import Path
from typing import List

import pytest
from nexusformat.nexus.tree import (
    NXentry,
    NXfield,
    NXgroup,
    NXinstrument,
    NXlog,
    NXtransformations,
    nxload,
)

from assonant.data_classes.components import Component
from assonant.data_classes.data_handlers import Axis, DataField, TimeSeries
from assonant.file_writer import AssonantFileWriter
from assonant.file_writer.exceptions import FileWriterFactoryError
from assonant.file_writer.file_writer_factory import FileWriterFactory

from ..utils import (
    assert_nxobject_and_assonant_axis_equality,
    assert_nxobject_and_assonant_data_field_equality,
    get_fixture_values,
    get_list_of_data_handlers_fixture_names,
)


@pytest.fixture(scope="session")
def nexus_file_writer() -> AssonantFileWriter:
    return AssonantFileWriter(file_format="nexus")


@pytest.fixture(scope="function")
def base_tmp_dir_path(tmp_path_factory) -> str:
    return str(tmp_path_factory.mktemp("test"))


def get_components_name_from_assonant_submodule() -> List[Component]:
    """
    Return a list with all names from existing Component classes within Components
    submodule from Assonant Data Classes.

    Returns:
        List[Component]: List with all existing Component class names within
        Components submodule from Assonant Data Classes.
    """
    return [
        member[0]
        for member in inspect.getmembers(importlib.import_module("assonant.data_classes.components"), inspect.isclass)
    ]


def test_invalid_file_format():
    """
    Validate that AssonantFileWriter constructor correctly raises exception
    if unsupported file_format is passed during object construction.

    Expected Behavior:
        - AssonantFileWriter must raise an FileWriterFactoryError if an unsupported file format is passed.
    """
    with pytest.raises(FileWriterFactoryError):
        _ = AssonantFileWriter(file_format="UNSUPPORTED_FILE_FORMAT")


def test_get_supported_file_formats():
    """
    Validate that AssonantFileWriter correctly return supported file formats.

    Expected Behavior:
        - AssonantFileWriter method must return a list of formats equal to the
        supported formats by the FileWriterFactory.
    """
    assert AssonantFileWriter.get_supported_file_formats() == FileWriterFactory.get_supported_file_formats()


def test_get_file_extension(nexus_file_writer):
    """
    Validate that AssonantFileWriter correctly return correct file extension.

    Expected Behavior:
        - AssonantFileWriter method must return the same file extension defined by the format specific
        file writer.
    """
    assert nexus_file_writer.get_file_extension() == nexus_file_writer.file_writer.get_file_extension()


@pytest.mark.parametrize(
    "component_name",
    [c_name for c_name in get_components_name_from_assonant_submodule() if c_name not in ["Beamline", "Component"]],
)
def test_write_component_as_nexus_group(
    nexus_file_writer, base_tmp_dir_path, example_components_collection, component_name, component_to_nexus_mapping
):
    """
    Validate that AssonantFileWriter correctly create NeXus groups related to passed
    Assonant data classes Component within NeXus files.

    PS: Beamline is a special case treated in a standalone test

    Expected Behavior:
        - AssonantFileWriter must create a new NeXus into target file path with the given file name
        - A group with the same name as the one given to the component must be created within the file
        - The group must be in the root level of the nexus hierarchy
    """
    filename = "test_nexus"
    empty_component = example_components_collection[component_name]
    component_nexus_type = component_to_nexus_mapping[type(empty_component)]
    group_name = component_name.lower()

    nexus_file_writer.write_data(filepath=base_tmp_dir_path, filename=filename, data=empty_component)

    output_file_path = Path(base_tmp_dir_path, filename).with_suffix(nexus_file_writer.get_file_extension())
    nxroot = nxload(output_file_path, mode="r")

    # Check group was created with component name and fits the correct NeXus object
    assert nxroot[group_name]
    assert isinstance(nxroot[group_name], NXgroup)
    assert nxroot[group_name].nxclass
    assert isinstance(nxroot[group_name], component_nexus_type)


def test_write_beamline_component_special_case_as_nexus_group(nexus_file_writer, base_tmp_dir_path, example_beamline):
    """
    Validate that AssonantFileWriter correctly write Assonant data classes Beamline specific case within NeXus files.

    Expected Behavior:
        - AssonantFileWriter must create a new NeXus into target file path with the given file name
        - A group with the name instrument must be created within the NeXus file
        - This group must be of NXinstrument type
        - This group must have a field called name with value equivalent to the Beamline name value
        - The group must be in the root level of the nexus hierarchy
    """
    filename = "test_nexus"
    empty_beamline = example_beamline

    nexus_file_writer.write_data(filepath=base_tmp_dir_path, filename=filename, data=empty_beamline)

    output_file_path = Path(base_tmp_dir_path, filename).with_suffix(nexus_file_writer.get_file_extension())
    nxroot = nxload(output_file_path, mode="r")

    # Check group was correctly created within NeXus file
    assert nxroot["instrument"]
    assert isinstance(nxroot["instrument"], NXinstrument)


@pytest.mark.parametrize(
    "example_data_field_fixture_name", get_list_of_data_handlers_fixture_names(data_handler_type=DataField)
)
def test_write_component_field_as_nexus_field(
    nexus_file_writer, base_tmp_dir_path, example_components_collection, example_data_field_fixture_name, request
):
    """
    Validate that AssonantFileWriter correctly write Assonant Component field as a NeXus field within its
    respective group in the NeXus file.

    Expected Behavior:
        - AssonantFileWriter must create a new NeXus into target file path with the given file name
        - A group with the same name as the device must be created
        - Inside the device group, data fields must exist with the respective given name with the same values
        they had before being written to the file.
    """
    filename = "test_nexus"
    component_name = "Mirror"
    component = example_components_collection[
        component_name
    ]  # Any Component fits this test as it isn't the tested featured.
    group_name = component_name.lower()

    example_data_field = get_fixture_values(example_data_field_fixture_name, request)

    component.add_field(example_data_field_fixture_name, example_data_field)

    nexus_file_writer.write_data(filepath=base_tmp_dir_path, filename=filename, data=component)

    output_file_path = Path(base_tmp_dir_path, filename).with_suffix(nexus_file_writer.get_file_extension())

    # Check file was created
    assert output_file_path.exists()

    nxroot = nxload(output_file_path, mode="r")

    # Check nexus content and structure is correct
    assert nxroot[group_name][example_data_field_fixture_name]
    assert isinstance(nxroot[group_name][example_data_field_fixture_name], NXfield)
    assert_nxobject_and_assonant_data_field_equality(
        nxroot[group_name][example_data_field_fixture_name], example_data_field
    )


@pytest.mark.parametrize(
    "example_data_field_fixture_name", get_list_of_data_handlers_fixture_names(data_handler_type=DataField)
)
def test_write_beamline_component_field_as_nexus_field(
    nexus_file_writer, base_tmp_dir_path, example_beamline, example_data_field_fixture_name, request
):
    """
    Validate that AssonantFileWriter correctly write Assonant Beamline Component field as a NeXus field within its
    respective group in the NeXus file.

    Expected Behavior:
        - AssonantFileWriter must create a new NeXus into target file path with the given file name
        - A group named instrument must be created to store Beamline information
        - Inside the instrument group, data fields must exist with the respective given name with the same values
        they had before being written to thefile.
    """
    filename = "test_nexus"
    component = example_beamline
    group_name = "instrument"  # Beamline specific case

    example_data_field = get_fixture_values(example_data_field_fixture_name, request)

    component.add_field(example_data_field_fixture_name, example_data_field)

    nexus_file_writer.write_data(filepath=base_tmp_dir_path, filename=filename, data=component)

    output_file_path = Path(base_tmp_dir_path, filename).with_suffix(nexus_file_writer.get_file_extension())

    # Check file was created
    assert output_file_path.exists()

    nxroot = nxload(output_file_path, mode="r")

    # Check nexus content and structure is correct
    assert nxroot[group_name][example_data_field_fixture_name]
    assert isinstance(nxroot[group_name][example_data_field_fixture_name], NXfield)
    assert_nxobject_and_assonant_data_field_equality(
        nxroot[group_name][example_data_field_fixture_name], example_data_field
    )


@pytest.mark.parametrize("example_axis_fixture_name", get_list_of_data_handlers_fixture_names(data_handler_type=Axis))
def test_write_component_position_as_nexus_field(
    nexus_file_writer, base_tmp_dir_path, example_components_collection, example_axis_fixture_name, request
):
    """
    Validate that AssonantFileWriter correctly write Assonant Component position as a NeXus field within its
    Component's respective NXtransformation group.

    Expected Behavior:
        - AssonantFileWriter must create a new NeXus into target file path with the given file name
        - A group with the same name as the device must be created
        - Inside the device group, a group called 'transformations' must be create
        - The 'transformations' group must be of NXtransformation type
        - Inside the 'transformations' group axis information must exist as NXfield with the respective given
          name with the same values they had before being written to the file.
    """
    filename = "test_nexus"
    component_name = "Mirror"
    component = example_components_collection[
        component_name
    ]  # Any Component fits this test as it isn't the tested featured.
    group_name = component_name.lower()

    example_axis = get_fixture_values(example_axis_fixture_name, request)

    component.add_position(example_axis_fixture_name, example_axis)

    nexus_file_writer.write_data(filepath=base_tmp_dir_path, filename=filename, data=component)

    output_file_path = Path(base_tmp_dir_path, filename).with_suffix(nexus_file_writer.get_file_extension())

    # Check file was created
    assert output_file_path.exists()

    nxroot = nxload(output_file_path, mode="r")

    # Check if transformation group was created
    assert nxroot[group_name]["transformations"]
    assert isinstance(nxroot[group_name]["transformations"], NXtransformations)

    # Check if transformation group content is correct
    assert nxroot[group_name]["transformations"][example_axis_fixture_name]
    assert isinstance(nxroot[group_name]["transformations"][example_axis_fixture_name], NXfield)
    assert_nxobject_and_assonant_axis_equality(
        nxroot[group_name]["transformations"][example_axis_fixture_name], example_axis
    )


# ===================== Warning =========================
# | TEST WRITE BEAMLINE POSITION WAS NOT IMPLEMENTED    |
# | AS THIS IS A UNDESIRED BEHAVIOR FUTURALLY, BEAMLINE |
# | COMPONENT MUST PROHIBIT ADDING POSITION TO IT AND A |
# | TEST TO VALIDATE IT MUST BE DONE                    |
# =======================================================


@pytest.mark.parametrize(
    "example_timeseries_fixture_name", get_list_of_data_handlers_fixture_names(data_handler_type=TimeSeries)
)
def test_write_component_timeseries_field_as_nexus_field(
    nexus_file_writer, base_tmp_dir_path, example_components_collection, example_timeseries_fixture_name, request
):
    """
    Validate that AssonantFileWriter correctly write Assonant Component timeseries as NeXus fields within its
    Component's respective group.

    Expected Behavior:
        - AssonantFileWriter must create a new NeXus into target file path with the given file name
        - A group with the same name as the device must be created
        - Inside the device group, a group with the name of the field must be create
        - The group with the field name must be of 'NXlog' type
        - Inside the group with the field name two NXfield must exist:
            - (1) 'value' field must store the same data as the TimeSeries value property
            - (2) 'time' field must store the same data as the TimeSeries timestamps property
        - If the Datahandler is actually an Axis with TimeSeries DataField, the same behavior
        is expected but inside the NXtransformation group
    """
    filename = "test_nexus"

    component_name = "Mirror"
    component = example_components_collection[
        component_name
    ]  # Any Component fits this test as it isn't the tested featured.
    group_name = component_name.lower()

    example_timeseries = get_fixture_values(example_timeseries_fixture_name, request)

    component.add_data_handler(example_timeseries_fixture_name, example_timeseries)

    nexus_file_writer.write_data(filepath=base_tmp_dir_path, filename=filename, data=component)

    output_file_path = Path(base_tmp_dir_path, filename).with_suffix(nexus_file_writer.get_file_extension())

    # Check file was created
    assert output_file_path.exists()

    nxroot = nxload(output_file_path, mode="r")

    # Check fixture tyṕe is correct to avoid bypassing test
    assert isinstance(example_timeseries, TimeSeries) or isinstance(example_timeseries, Axis)

    # Check nexus content and structure is correct relative to the time of TimeSeries
    if isinstance(example_timeseries, TimeSeries):
        timeseries_path = "/".join([group_name, example_timeseries_fixture_name])
        assertion_method = assert_nxobject_and_assonant_data_field_equality
        relative_data_field_ref = example_timeseries.value
        relative_timestamps_ref = example_timeseries.timestamps
    elif isinstance(example_timeseries, Axis):
        timeseries_path = "/".join([group_name, "transformations", example_timeseries_fixture_name])
        assertion_method = assert_nxobject_and_assonant_axis_equality
        relative_data_field_ref = example_timeseries
        relative_timestamps_ref = example_timeseries.value.timestamps

    assert nxroot[timeseries_path]
    assert isinstance(nxroot[timeseries_path], NXlog)
    assert nxroot[timeseries_path]["value"]
    assert isinstance(nxroot[timeseries_path]["value"], NXfield)
    assert nxroot[timeseries_path]["time"]
    assert isinstance(nxroot[timeseries_path]["time"], NXfield)

    assertion_method(nxroot[timeseries_path]["value"], relative_data_field_ref)
    assert_nxobject_and_assonant_data_field_equality(nxroot[timeseries_path]["time"], relative_timestamps_ref)


@pytest.mark.parametrize(
    "subcomponent_name",
    [c_name for c_name in get_components_name_from_assonant_submodule() if c_name not in ["Beamline", "Component"]],
)
def test_write_subcomponent_as_nexus_group(
    nexus_file_writer, base_tmp_dir_path, example_components_collection, subcomponent_name, component_to_nexus_mapping
):
    """
    Validate that AssonantFileWriter correctly create NeXus groups related to passed
    Assonant data classes Component and its subcomponent within NeXus files.

    PS: Beamline is a special case treated in a standalone test

    Expected Behavior:
        - AssonantFileWriter must create a new NeXus into target file path with the given file name
        - A group with the same name as the one given to the component must be created within the file
        - A group within the component group must be created with the name name as the subcomponent
        - The subgroup nexus class must fit the respective NeXus class for the subcomponent type
    """
    filename = "test_nexus"
    component_name = "Monochromator"  # Don't matter which component is used as subcomponents are the focus
    empty_component = example_components_collection[component_name].__deepcopy__()  # deepcopy to avoid infinite loop
    empty_subcomponent = example_components_collection[subcomponent_name]
    subcomponent_nexus_type = component_to_nexus_mapping[type(empty_subcomponent)]
    component_group_name = component_name.lower()
    subcomponent_group_name = subcomponent_name.lower()

    empty_component.add_subcomponent(empty_subcomponent)

    nexus_file_writer.write_data(filepath=base_tmp_dir_path, filename=filename, data=empty_component)

    output_file_path = Path(base_tmp_dir_path, filename).with_suffix(nexus_file_writer.get_file_extension())
    nxroot = nxload(output_file_path, mode="r")

    # Check subgroup was created with subcomponent name and with correct Nexus type
    assert nxroot[component_group_name][subcomponent_group_name]
    assert isinstance(nxroot[component_group_name][subcomponent_group_name], NXgroup)
    assert nxroot[component_group_name][subcomponent_group_name].nxclass
    assert isinstance(nxroot[component_group_name][subcomponent_group_name], subcomponent_nexus_type)


def test_write_beamline_subcomponent_as_nexus_group(
    nexus_file_writer, base_tmp_dir_path, example_components_collection
):
    """
    Validate that AssonantFileWriter correctly create NeXus groups related to passed
    Assonant data classes Beamline Component and its subcomponent within NeXus files.

    PS: No need to test all subcomponent types as they were already tested on previous test

    Expected Behavior:
        - AssonantFileWriter must create a new NeXus into target file path with the given file name
        - A group with the name 'instrument' must be created within the file for the Beamline Component
        - A group within the 'instrument' group must be created with the name as the subcomponent
    """
    filename = "test_nexus"
    component_name = "Beamline"
    subcomponent_name = "Mirror"
    empty_component = example_components_collection[component_name]
    empty_subcomponent = example_components_collection[subcomponent_name]
    component_group_name = "instrument"
    subcomponent_group_name = subcomponent_name.lower()

    empty_component.add_subcomponent(empty_subcomponent)

    nexus_file_writer.write_data(filepath=base_tmp_dir_path, filename=filename, data=empty_component)

    output_file_path = Path(base_tmp_dir_path, filename).with_suffix(nexus_file_writer.get_file_extension())
    nxroot = nxload(output_file_path, mode="r")

    # Check subgroup was created with subcomponent name and with correct Nexus type
    assert nxroot[component_group_name][subcomponent_group_name]
    assert isinstance(nxroot[component_group_name][subcomponent_group_name], NXgroup)
    assert nxroot[component_group_name][subcomponent_group_name].nxclass


def test_write_data_field_within_subcomponent_as_nexus_field(
    nexus_file_writer, base_tmp_dir_path, example_components_collection, example_int_1d_array_data_field
):
    """
    Validate that AssonantFileWriter correctly save subcomponent data field within NeXus file

    PS: Don't need to extensively test data types as that was already tested on previous tests.

    Expected Behavior:
        - AssonantFileWriter must create a new NeXus into target file path with the given file name
        - A group with the same name as the one given to the component must be created within the file
        - A group within the component group must be created with the name as the subcomponent
        - A nexus field must exist within the subcomponent group with its content equivalent
        to Assonant DataField
    """
    filename = "test_nexus"
    component_name = "Monochromator"  # Don't matter which componet is used as subcomponents are the focus
    subcomponent_name = "Mirror"
    data_field_name = "data_field"
    component = example_components_collection[component_name]
    subcomponent = example_components_collection[subcomponent_name]
    component_group_name = component_name.lower()
    subcomponent_group_name = subcomponent_name.lower()

    subcomponent.add_field(data_field_name, example_int_1d_array_data_field)
    component.add_subcomponent(subcomponent)

    nexus_file_writer.write_data(filepath=base_tmp_dir_path, filename=filename, data=component)

    output_file_path = Path(base_tmp_dir_path, filename).with_suffix(nexus_file_writer.get_file_extension())
    nxroot = nxload(output_file_path, mode="r")

    path_to_nxfield = "/".join([component_group_name, subcomponent_group_name, data_field_name])

    # Check subgroup was created with subcomponent name and with correct Nexus type
    assert nxroot[path_to_nxfield]
    assert isinstance(nxroot[path_to_nxfield], NXfield)
    assert_nxobject_and_assonant_data_field_equality(nxroot[path_to_nxfield], example_int_1d_array_data_field)


def test_write_axis_within_subcomponent_as_nexus_field(
    nexus_file_writer, base_tmp_dir_path, example_components_collection, example_translation_axis
):
    """
    Validate that AssonantFileWriter correctly save subcomponent axis within NeXus file

    PS: Don't need to extensively test data types as that was already tested on previous tests.

    Expected Behavior:
        - AssonantFileWriter must create a new NeXus into target file path with the given file name
        - A group with the same name as the one given to the component must be created within the file
        - A group within the component group must be created with the name as the subcomponent
        - A group within the subcomponent group named 'transformations' must be created
        - A nexus field must exist within the 'transformation' group with its name and content equivalent
        to Assonant Axis
    """
    filename = "test_nexus"
    component_name = "Monochromator"  # Don't matter which componet is used as subcomponents are the focus
    subcomponent_name = "Mirror"
    axis_name = "axis_name"
    component = example_components_collection[component_name]
    subcomponent = example_components_collection[subcomponent_name]
    component_group_name = component_name.lower()
    subcomponent_group_name = subcomponent_name.lower()

    subcomponent.add_position(axis_name, example_translation_axis)
    component.add_subcomponent(subcomponent)

    nexus_file_writer.write_data(filepath=base_tmp_dir_path, filename=filename, data=component)

    output_file_path = Path(base_tmp_dir_path, filename).with_suffix(nexus_file_writer.get_file_extension())
    nxroot = nxload(output_file_path, mode="r")

    path_to_nxfield = "/".join([component_group_name, subcomponent_group_name, "transformations", axis_name])

    # Check subgroup was created with subcomponent name and with correct Nexus type
    assert nxroot[path_to_nxfield]
    assert isinstance(nxroot[path_to_nxfield], NXfield)
    assert_nxobject_and_assonant_axis_equality(nxroot[path_to_nxfield], example_translation_axis)


def test_write_data_field_timeseries_within_subcomponent_as_nexus_field(
    nexus_file_writer, base_tmp_dir_path, example_components_collection, example_int_1d_timeseries_data_field
):
    """
    Validate that AssonantFileWriter correctly save TimeSeries DataField within NeXus file

    PS: Don't need to extensively test data types as that was already tested on previous tests.

    Expected Behavior:
        - AssonantFileWriter must create a new NeXus into target file path with the given file name
        - A group with the same name as the one given to the component must be created within the file
        - A group within the component group must be created with the name as the subcomponent
        - A group within the subcomponent group with the same name as the TimeSeries DataHandler must be created
        - The TimeSeries DataHandler group must be of type NXlog
        - Two nexus fields must exist within the TimeSeries group:
            (1) 'value': nexus field storing value property from the TimeSeries
            (2) 'time': nexus field storing timestamps property from the TimeSeries]
        - The TimeSeries group content must be equivalent to Assonant TimeSeries DataHandler
    """
    filename = "test_nexus"
    component_name = "Monochromator"  # Don't matter which componet is used as subcomponents are the focus
    subcomponent_name = "Mirror"
    timeseries_data_field_name = "timeseries_data_field"
    component = example_components_collection[component_name]
    subcomponent = example_components_collection[subcomponent_name]
    component_group_name = component_name.lower()
    subcomponent_group_name = subcomponent_name.lower()

    subcomponent.add_timeseries_field(timeseries_data_field_name, example_int_1d_timeseries_data_field)
    component.add_subcomponent(subcomponent)

    nexus_file_writer.write_data(filepath=base_tmp_dir_path, filename=filename, data=component)

    output_file_path = Path(base_tmp_dir_path, filename).with_suffix(nexus_file_writer.get_file_extension())
    nxroot = nxload(output_file_path, mode="r")

    path_to_value_nxfield = "/".join(
        [component_group_name, subcomponent_group_name, timeseries_data_field_name, "value"]
    )
    path_to_time_nxfield = "/".join([component_group_name, subcomponent_group_name, timeseries_data_field_name, "time"])

    # Check subgroup was created with subcomponent name and with correct Nexus type
    assert nxroot[path_to_value_nxfield]
    assert nxroot[path_to_time_nxfield]
    assert isinstance(nxroot[path_to_value_nxfield], NXfield)
    assert isinstance(nxroot[path_to_time_nxfield], NXfield)
    assert_nxobject_and_assonant_data_field_equality(
        nxroot[path_to_value_nxfield], example_int_1d_timeseries_data_field.value
    )
    assert_nxobject_and_assonant_data_field_equality(
        nxroot[path_to_time_nxfield], example_int_1d_timeseries_data_field.timestamps
    )


def test_write_axis_timeseries_within_subcomponent_as_nexus_field(
    nexus_file_writer, base_tmp_dir_path, example_components_collection, example_timeseries_rotation_axis
):
    """
    Validate that AssonantFileWriter correctly save Axis TimeSeries within NeXus file

    PS: Don't need to extensively test data types as that was already tested on previous tests.

    Expected Behavior:
        - AssonantFileWriter must create a new NeXus into target file path with the given file name
        - A group with the same name as the one given to the component must be created within the file
        - A group within the component group must be created with the name as the subcomponent
        - A group within the subcomponent group named 'transformations' must exist
        - A group within the 'transformations' group with the same name as the TimeSeries DataHandler must be created
        - The TimeSeries DataHandler group must be of type NXlog
        - Two nexus fields must exist within the TimeSeries group:
            (1) 'value': nexus field storing value property from the TimeSeries
            (2) 'time': nexus field storing timestamps property from the TimeSeries]
        - The Axis TimeSeries group content must be equivalent to Assonant Axis TimeSeries DataHandler
    """
    filename = "test_nexus"
    component_name = "Monochromator"  # Don't matter which componet is used as subcomponents are the focus
    subcomponent_name = "Mirror"
    timeseries_axis_name = "timeseries_axis"
    component = example_components_collection[component_name]
    subcomponent = example_components_collection[subcomponent_name]
    component_group_name = component_name.lower()
    subcomponent_group_name = subcomponent_name.lower()

    subcomponent.add_position(timeseries_axis_name, example_timeseries_rotation_axis)
    component.add_subcomponent(subcomponent)

    nexus_file_writer.write_data(filepath=base_tmp_dir_path, filename=filename, data=component)

    output_file_path = Path(base_tmp_dir_path, filename).with_suffix(nexus_file_writer.get_file_extension())
    nxroot = nxload(output_file_path, mode="r")

    path_to_axis_group = "/".join(
        [component_group_name, subcomponent_group_name, "transformations", timeseries_axis_name]
    )

    # Check subgroup was created with subcomponent name and with correct Nexus type
    assert nxroot[path_to_axis_group]["value"]
    assert nxroot[path_to_axis_group]["time"]
    assert isinstance(nxroot[path_to_axis_group]["value"], NXfield)
    assert isinstance(nxroot[path_to_axis_group]["time"], NXfield)
    assert_nxobject_and_assonant_axis_equality(nxroot[path_to_axis_group]["value"], example_timeseries_rotation_axis)
    assert_nxobject_and_assonant_data_field_equality(
        nxroot[path_to_axis_group]["time"], example_timeseries_rotation_axis.value.timestamps
    )


def test_write_entries_without_subentries_with_name_conflict_in_nexus(
    nexus_file_writer,
    base_tmp_dir_path,
    example_entry,
    example_int_1d_array_data_field,
    example_float_2d_array_data_field,
):
    """
    Validate that AssonantFileWriter correctly handle writing entries with equal name in the nexus file

    Expected Behavior:
        - AssonantFileWriter must create a new NeXus into target file path with the given file name
        - Due to entry name conflict, a new file must be created to avoid data loss
        - The new file name must be equal to the original name with a index posfix (e.g: 000001)
        - Inside both files, name and content of both entries must be equivalent to its objects data
    """
    filename = "test_nexus"
    entry_name = example_entry.name
    data_field_name = "data_field"

    entry_1 = example_entry.model_copy(deep=True)
    entry_2 = example_entry.model_copy(deep=True)

    entry_1.add_field(data_field_name, example_int_1d_array_data_field)
    entry_2.add_field(data_field_name, example_float_2d_array_data_field)

    nexus_file_writer.write_data(filepath=base_tmp_dir_path, filename=filename, data=entry_1)

    nexus_file_writer.write_data(filepath=base_tmp_dir_path, filename=filename, data=entry_2)

    output_file_1_path = Path(base_tmp_dir_path, filename).with_suffix(nexus_file_writer.get_file_extension())
    output_file_2_path = Path(base_tmp_dir_path, filename + "_000001").with_suffix(
        nexus_file_writer.get_file_extension()
    )

    # Check file was created
    assert output_file_1_path.exists()
    assert output_file_2_path.exists()

    nxroot_file_1 = nxload(output_file_1_path, mode="r")
    nxroot_file_2 = nxload(output_file_2_path, mode="r")

    # Check nexus content and structure is correct in both files
    assert nxroot_file_1[entry_name]
    assert isinstance(nxroot_file_1[entry_name], NXentry)
    assert nxroot_file_1[entry_name][data_field_name]
    assert isinstance(nxroot_file_1[entry_name][data_field_name], NXfield)
    assert_nxobject_and_assonant_data_field_equality(
        nxroot_file_1[entry_name][data_field_name], example_int_1d_array_data_field
    )

    assert nxroot_file_2[entry_name]
    assert isinstance(nxroot_file_2[entry_name], NXentry)
    assert nxroot_file_2[entry_name][data_field_name]
    assert isinstance(nxroot_file_2[entry_name][data_field_name], NXfield)
    assert_nxobject_and_assonant_data_field_equality(
        nxroot_file_2[entry_name][data_field_name], example_float_2d_array_data_field
    )


def test_write_entries_without_conflict_in_nexus(
    nexus_file_writer, base_tmp_dir_path, example_entry, example_int_1d_array_data_field
):
    """
    Validate that AssonantFileWriter correctly handle writing entries without conflicting name in the nexus file

    Expected Behavior:
        - AssonantFileWriter must create a new NeXus into target file path with the given file name
        - As no name conflict exists between entries, both entries must exist within the same file
        - One NXentry group must exist for each entry with the name given to the entry
        - Both entry content must be equivalent to how it was represented before saved in NeXus
    """
    filename = "test_nexus"
    data_field_name = "data_field"

    entry_1 = example_entry.model_copy(deep=True)
    entry_2 = example_entry.model_copy(deep=True)

    # Differ entry names for test
    entry_1.name = "entry_1_name"
    entry_2.name = "entry_2_name"

    entry_1.add_field(data_field_name, example_int_1d_array_data_field)
    entry_2.add_field(data_field_name, example_int_1d_array_data_field)

    nexus_file_writer.write_data(filepath=base_tmp_dir_path, filename=filename, data=entry_1)

    nexus_file_writer.write_data(filepath=base_tmp_dir_path, filename=filename, data=entry_2)

    output_file_1_path = Path(base_tmp_dir_path, filename).with_suffix(nexus_file_writer.get_file_extension())

    # Check file was created
    assert output_file_1_path.exists()
    nx_root = nxload(output_file_1_path, mode="r")

    # Check nexus content and structure is correct in both files
    assert nx_root[entry_1.name]
    assert isinstance(nx_root[entry_1.name], NXentry)
    assert nx_root[entry_1.name][data_field_name]
    assert isinstance(nx_root[entry_1.name][data_field_name], NXfield)
    assert_nxobject_and_assonant_data_field_equality(
        nx_root[entry_1.name][data_field_name], example_int_1d_array_data_field
    )

    assert nx_root[entry_2.name]
    assert isinstance(nx_root[entry_2.name], NXentry)
    assert nx_root[entry_2.name][data_field_name]
    assert isinstance(nx_root[entry_2.name][data_field_name], NXfield)
    assert_nxobject_and_assonant_data_field_equality(
        nx_root[entry_2.name][data_field_name], example_int_1d_array_data_field
    )


def test_write_subentries_without_conflict_within_conflicting_entry_in_nexus(
    nexus_file_writer, base_tmp_dir_path, example_entry, example_int_1d_array_data_field
):
    """
    Validate that AssonantFileWriter correctly handle writing subentries without conflict name within
    the same entry in nexus file.

    Expected Behavior:
        - AssonantFileWriter must create a new NeXus into target file path with the given file name
        - As a name conflict exists between entries, conflict between its subcomponents must be checked
        - As both suybcomponents are the subetris and they don't have conflicting names, they can be written
        within the same file under the same entry
        - One NXentry group must exist for the entry with the name given to the entry
        - Both subentries must be found within the entry and its content must be equivalent to how it was
          represented before saved in NeXus
    """
    filename = "test_nexus"
    data_field_name = "data_field"

    entry_1 = example_entry.model_copy(deep=True)
    entry_2 = example_entry.model_copy(deep=True)
    subentry_1 = example_entry.model_copy(deep=True)
    subentry_2 = example_entry.model_copy(deep=True)

    # Keep entry names equal for test just differ subentry names
    subentry_1.name = "subentry_1"
    subentry_2.name = "subentry_2"
    subentry_1.add_field(data_field_name, example_int_1d_array_data_field)
    subentry_2.add_field(data_field_name, example_int_1d_array_data_field)

    entry_1.add_subcomponent(subentry_1)
    entry_2.add_subcomponent(subentry_2)

    nexus_file_writer.write_data(filepath=base_tmp_dir_path, filename=filename, data=entry_1)
    nexus_file_writer.write_data(filepath=base_tmp_dir_path, filename=filename, data=entry_2)

    output_file_1_path = Path(base_tmp_dir_path, filename).with_suffix(nexus_file_writer.get_file_extension())

    # Check file was created
    assert output_file_1_path.exists()
    nx_root = nxload(output_file_1_path, mode="r")

    # Check nexus content and structure is correct in both files
    assert nx_root[entry_1.name]
    assert isinstance(nx_root[entry_1.name], NXentry)
    assert nx_root[entry_1.name][subentry_1.name]
    assert isinstance(nx_root[entry_1.name][subentry_1.name], NXentry)
    assert nx_root[entry_1.name][subentry_1.name][data_field_name]
    assert isinstance(nx_root[entry_1.name][subentry_1.name][data_field_name], NXfield)
    assert_nxobject_and_assonant_data_field_equality(
        nx_root[entry_1.name][subentry_1.name][data_field_name], example_int_1d_array_data_field
    )
    assert nx_root[entry_1.name][subentry_2.name]
    assert isinstance(nx_root[entry_1.name][subentry_2.name], NXentry)
    assert nx_root[entry_1.name][subentry_2.name][data_field_name]
    assert isinstance(nx_root[entry_1.name][subentry_2.name][data_field_name], NXfield)
    assert_nxobject_and_assonant_data_field_equality(
        nx_root[entry_1.name][subentry_2.name][data_field_name], example_int_1d_array_data_field
    )


def test_write_subentries_with_conflicting_entry_and_subentry_but_different_data_fields(
    nexus_file_writer,
    base_tmp_dir_path,
    example_entry,
    example_int_1d_array_data_field,
    example_float_2d_array_data_field,
):
    """
    Validate that AssonantFileWriter correctly handle writing data_fields to conflicting entry and
    subentries when data_field names don't conflict.

    Expected Behavior:
        - AssonantFileWriter must create a new NeXus into target file path with the given file name
        - As a name conflict exists between entries, conflict between its subcomponents must be checked
        - As both subcomponents are the subentries and they have conflicting names, conflict between its
            subcomponents must be checked.
        - As no conflict exist between data_field names, both must be written within the same subentry.
        - One NXentry group must exist for the entry with the name given to the entry
        - Within that entry one Subentry must exist with the name given to the subentry
        - Both data_fields must be found within the subentry and its content must be equivalent to how it was
          represented before saved in NeXus
    """
    filename = "test_nexus"
    data_field_1_name = "data_field_1"
    data_field_2_name = "data_field_2"

    entry_1 = example_entry.model_copy(deep=True)
    entry_2 = example_entry.model_copy(deep=True)
    subentry_1 = example_entry.model_copy(deep=True)
    subentry_2 = example_entry.model_copy(deep=True)

    # Keep entry names equal for test just differ subentry names
    subentry_1.name = "subentry"
    subentry_2.name = "subentry"
    subentry_1.add_field(data_field_1_name, example_int_1d_array_data_field)
    subentry_2.add_field(data_field_2_name, example_float_2d_array_data_field)

    entry_1.add_subcomponent(subentry_1)
    entry_2.add_subcomponent(subentry_2)

    nexus_file_writer.write_data(filepath=base_tmp_dir_path, filename=filename, data=entry_1)
    nexus_file_writer.write_data(filepath=base_tmp_dir_path, filename=filename, data=entry_2)

    output_file_1_path = Path(base_tmp_dir_path, filename).with_suffix(nexus_file_writer.get_file_extension())

    # Check file was created
    assert output_file_1_path.exists()
    nx_root = nxload(output_file_1_path, mode="r")

    # Check nexus content and structure is correct in both files
    assert nx_root[entry_1.name]
    assert isinstance(nx_root[entry_1.name], NXentry)
    assert nx_root[entry_1.name][subentry_1.name]
    assert isinstance(nx_root[entry_1.name][subentry_1.name], NXentry)
    assert nx_root[entry_1.name][subentry_1.name][data_field_1_name]
    assert isinstance(nx_root[entry_1.name][subentry_1.name][data_field_1_name], NXfield)
    assert_nxobject_and_assonant_data_field_equality(
        nx_root[entry_1.name][subentry_1.name][data_field_1_name], example_int_1d_array_data_field
    )
    assert nx_root[entry_1.name][subentry_1.name][data_field_2_name]
    assert isinstance(nx_root[entry_1.name][subentry_1.name][data_field_2_name], NXfield)
    assert_nxobject_and_assonant_data_field_equality(
        nx_root[entry_1.name][subentry_1.name][data_field_2_name], example_float_2d_array_data_field
    )


def test_write_subentries_with_conflicting_entry_subentry_and_data_fields(
    nexus_file_writer,
    base_tmp_dir_path,
    example_entry,
    example_int_1d_array_data_field,
    example_float_2d_array_data_field,
):
    """
    Validate that AssonantFileWriter correctly handle writing data_fields to conflicting entry and
    subentries when data_field names don't conflict.

    Expected Behavior:
        - AssonantFileWriter must create a new NeXus into target file path with the given file name
        - As a name conflict exists between entries, conflict between its subcomponents must be checked
        - As both subcomponents are the subentries and they have conflicting names, conflict between its
            subcomponents must be checked.
        - As data_fields have conflicting names and they are Non-Entry Components, storing the second data_field
            must fail
        - An Exception of type FileAlreadyExistsError must be raised.
        - A new file with the original name + a posfix (e.g: 000001) must be created.
        - Within the new file a mirrored structure from the original file must exist but with the second data_field
          within the subentry.

    """
    filename = "test_nexus"
    data_field_name = "data_field"

    entry_1 = example_entry.model_copy(deep=True)
    entry_2 = example_entry.model_copy(deep=True)
    subentry_1 = example_entry.model_copy(deep=True)
    subentry_2 = example_entry.model_copy(deep=True)

    # Keep entry names equal for test just differ subentry names
    subentry_1.name = "subentry"
    subentry_2.name = "subentry"
    subentry_1.add_field(data_field_name, example_int_1d_array_data_field)
    subentry_2.add_field(data_field_name, example_float_2d_array_data_field)

    entry_1.add_subcomponent(subentry_1)
    entry_2.add_subcomponent(subentry_2)

    nexus_file_writer.write_data(filepath=base_tmp_dir_path, filename=filename, data=entry_1)
    nexus_file_writer.write_data(filepath=base_tmp_dir_path, filename=filename, data=entry_2)

    output_file_1_path = Path(base_tmp_dir_path, filename).with_suffix(nexus_file_writer.get_file_extension())
    output_file_2_path = Path(base_tmp_dir_path, filename + "_000001").with_suffix(
        nexus_file_writer.get_file_extension()
    )

    # Check file was created
    assert output_file_1_path.exists()
    assert output_file_2_path.exists()
    nx_root = nxload(output_file_1_path, mode="r")

    # Check nexus content and structure is correct in both files
    assert nx_root[entry_1.name]
    assert isinstance(nx_root[entry_1.name], NXentry)
    assert nx_root[entry_1.name][subentry_1.name]
    assert isinstance(nx_root[entry_1.name][subentry_1.name], NXentry)
    assert nx_root[entry_1.name][subentry_1.name][data_field_name]
    assert isinstance(nx_root[entry_1.name][subentry_1.name][data_field_name], NXfield)
    assert_nxobject_and_assonant_data_field_equality(
        nx_root[entry_1.name][subentry_1.name][data_field_name], example_int_1d_array_data_field
    )

    nx_root = nxload(output_file_2_path, mode="r")

    # As structure MUST be mirrored fron the original file, entry_1 and subentry_1 components
    # are being used here.
    assert nx_root[entry_1.name]
    assert isinstance(nx_root[entry_1.name], NXentry)
    assert nx_root[entry_1.name][subentry_1.name]
    assert isinstance(nx_root[entry_1.name][subentry_1.name], NXentry)
    assert nx_root[entry_1.name][subentry_1.name][data_field_name]
    assert isinstance(nx_root[entry_1.name][subentry_1.name][data_field_name], NXfield)
    assert_nxobject_and_assonant_data_field_equality(
        nx_root[entry_1.name][subentry_1.name][data_field_name], example_float_2d_array_data_field
    )
    assert nx_root[entry_1.name][subentry_1.name][data_field_name]
    assert isinstance(nx_root[entry_1.name][subentry_1.name][data_field_name], NXfield)
    assert_nxobject_and_assonant_data_field_equality(
        nx_root[entry_1.name][subentry_1.name][data_field_name], example_float_2d_array_data_field
    )


def test_write_entry_with_component_name_conflict_in_nexus(
    nexus_file_writer, base_tmp_dir_path, example_entry, example_components_collection
):
    """
    Validate that AssonantFileWriter correctly handle writing Components with equal name in nexus file

    Expected Behavior:
        - AssonantFileWriter must create a new NeXus into target file path with the given file name
        - Both groups must be store within the same file
        - Due to name conflict, to avoid data loss/overwrite, the second group must be renamed
        - The new name must be the original group name with a posfix index (e.g:' (1)')
        - The content of both groups must be equivalent to how it was before being stored within file
    """
    filename = "test_nexus"
    component_name = "Mirror"
    component_1 = example_components_collection[component_name].model_copy(deep=True)
    component_2 = example_components_collection[component_name].model_copy(deep=True)
    component_group_name = component_name.lower()

    example_entry.add_subcomponent([component_1, component_2])

    nexus_file_writer.write_data(filepath=base_tmp_dir_path, filename=filename, data=example_entry)

    output_file_path = Path(base_tmp_dir_path, filename).with_suffix(nexus_file_writer.get_file_extension())
    nxroot = nxload(output_file_path, mode="r")

    # Check group was created with component name
    assert nxroot[example_entry.name][component_group_name]
    assert nxroot[example_entry.name][component_group_name + " (1)"]


def test_write_component_with_field_name_conflict_in_nexus(
    nexus_file_writer,
    base_tmp_dir_path,
    example_components_collection,
    example_int_1d_array_data_field,
    example_float_2d_array_data_field,
):
    """
    Validate that AssonantFileWriter correctly handle writing fields with equal name in nexus file

    Expected Behavior:
        - AssonantFileWriter must create a new NeXus into target file path with the given file name
        - Both fields must be created within the same Component group
        - Due to name conflict, to avoid data loss/overwrite, the second field must be renamed
        - The new name must be the original field name with a posfix index (e.g:' (1)')
        - The content of both groups must be equivalent to how it was before being stored within file
    """
    filename = "test_nexus"
    component_name = "Mirror"
    field_name = "data_field_name"
    component = example_components_collection[component_name]
    component_group_name = component_name.lower()

    component.add_field(field_name, example_int_1d_array_data_field)
    component.add_field(field_name, example_float_2d_array_data_field)

    nexus_file_writer.write_data(filepath=base_tmp_dir_path, filename=filename, data=component)

    output_file_path = Path(base_tmp_dir_path, filename).with_suffix(nexus_file_writer.get_file_extension())
    nxroot = nxload(output_file_path, mode="r")

    # Check group was created with component name
    assert nxroot[component_group_name][field_name]
    assert nxroot[component_group_name][field_name + " (1)"]

    assert_nxobject_and_assonant_data_field_equality(
        nxroot[component_group_name][field_name], example_int_1d_array_data_field
    )
    assert_nxobject_and_assonant_data_field_equality(
        nxroot[component_group_name][field_name + " (1)"], example_float_2d_array_data_field
    )



from typing import Any, Dict, List, Union

from assonant._bluesky import DataParserFactory
from assonant.naming_standards import AcquisitionType

from .assonant_data_retriever_interface import IAssonantDataRetriever
from .exceptions import AssonantDataRetrieverError


class BlueskyDataRetriever(IAssonantDataRetriever):
    """Bluesky Data Retriever."""

    retrieved_data = None

    @property
    def bluesky_data_parser(self):
        """Bluesky Data Parser getter."""
        return self._bluesky_data_parser

    @bluesky_data_parser.setter
    def bluesky_data_parser(self, parser):
        """Bluesky Data Parser setter. Invalidzgds retrieved data cache when set"""
        self._bluesky_data_parser = parser
        # Always invalidate cache when parser changes
        self.retrieved_data = None

    def __init__(self, bluesky_data_source: Union[List[List[Union[str, Dict[str, Any]]]], str]):
        """Instantiates BlueskyDataRetriever to retrieve data from given data source.

        Args:
            bluesky_data_source (Union[List[List[Union[str, Dict[str, Any]]]], str]): DataSource containing
            the bluesky data. Currently, supported formats are: JSON file containing bluesky events or a
            List of bluesky documents data presented also as  a list. The list with the document data has
            2 positions, the 1st is the bluesky identification for the document
            (start, event, stop, descriptor, ...) and the 2nd position the dict containing the document
            data.

        Raises:
            AssonantDataRetrieverError: Failed to create the suitable Bluesky Data Parser for given data source.
        """
        try:
            self.bluesky_data_parser = DataParserFactory.create_data_parser(bluesky_data_source=bluesky_data_source)
        except Exception as e:
            raise AssonantDataRetrieverError(
                "BlueskyDataRetriever failed to create Bluesky DataParser for given data source."
            ) from e

    def _convert_acquisition_type_into_event_stream_name(self, acquisition_type: AcquisitionType) -> List[str]:
        """Convert AcquisitionType value to its equivalent EventStreamName objects.

        PS: On Bluesky, the existing DataStreams can be mapped to a respective AcquisitionType. Current mapping is:
            * SNAPSHOT and VARY-> 'baseline'
            * SCAN -> All streams except 'baseline'

        Args:
            acquisition_type (AcquisitionType): AcquisitionType Enum object to be converted

        Raises:
            AssonantDataRetrieverError: Raised if there is no valid convertion for the passed AcquisitionType object.

        Returns:
            List[str]: List of stream names related to the passed Acquisition Type Enum object.
        """
        if acquisition_type in [AcquisitionType.SNAPSHOT_AT_START, AcquisitionType.SNAPSHOT_AT_END]:
            return ["baseline"]
        elif acquisition_type == AcquisitionType.SCAN:
            stream_names = self.bluesky_data_parser.get_stream_names()
            if "baseline" in stream_names:
                stream_names.remove("baseline")
            return stream_names  # Except baseline stream, all other streams fits SCAN AcquisitionType.
        else:
            raise AssonantDataRetrieverError(
                f"There is no valid EventStream convertion value for '{acquisition_type.value}' AcquisitionType"
            )

    def _merge_by_seq_num(
        self, input_dict: Dict[int, Dict[str, List[Any]]]
    ) -> Dict[str, Dict[str, Union[Any, List[Any]]]]:
        """Merge retrieved data by sequence number.

        Output structure:
        {
            PV_NAME_1: {
                'data': {
                    'value': [PV_VALUE_SEQ_NUM_1, PV_VALUE_SEQ_NUM_2, ...]
                    'timestamps': [PV_TIMESTAMP_SEQ_NUM_1, PV_TIMESTAMP_SEQ_NUM_2, ...]
                },
                'configs': PV_CONFIGS
            },
            ...
        }

        Args:
            input_dict (Dict[int, Dict[str, List[Any]]]): Dict returned by BlueskyDataParser
            with data stored by sequence number and PV name. Check BlueskyDataParser
            for details on the structure of the input dict.

        Returns:
            Dict[str, Dict[str, Union[Any, List[Any]]]]: Dict with information merged
            as the output structure described above.
        """
        merged = {}

        # Sort sequences
        for seq_num in sorted(input_dict.keys()):
            seq_data = input_dict[seq_num]

            for pv_name, pv_info in seq_data.items():
                value = pv_info["data"]["value"]
                timestamp = pv_info["data"]["timestamps"]
                configs = pv_info["configs"]

                if pv_name not in merged:
                    merged[pv_name] = {"data": {"value": [], "timestamps": []}, "configs": configs}

                merged[pv_name]["data"]["value"].append(value)
                merged[pv_name]["data"]["timestamps"].append(timestamp)

        return merged

    def get_pv_data_by_acquisition_type(self, acquisition_type: AcquisitionType) -> Dict[str, Union[Any, List[Any]]]:
        """Return collected PV data related to specified AcquisitionType.

        The returned value is a dictionary as follow:

        {
            PV_NAME_1: [PV_VALUE_1, PV_VALUE_2, ...],
            PV_NAME_2: [PV_VALUE_1],
            ...
        }

        PS: PV values are always returned within a list, even if it is a single value.

        Args:
            acquisition_type (AcquisitionType): Target Acquisition Type used to select
            data from which Acquisition Type will be retrieved.

        Returns:
            Dict[str, Union[Any, List[Any]]]: Dictionary containing retrieved PV data.
        """
        event_stream_names = self._convert_acquisition_type_into_event_stream_name(acquisition_type)
        data_retrived_by_query = {}

        # Cache retrieved data to avoid multiple retrievals from the same data source
        if self.retrieved_data is None:
            self.retrieved_data = self.bluesky_data_parser.query_all_data(with_config=True)

        # Filter data to fit the requested streams based on AcquisitionType
        for event_stream_name in event_stream_names:

            retrieved_stream_data = self.retrieved_data.get(event_stream_name, {event_stream_name: {}})
            # Special processing for SNAPSHOT cases
            if event_stream_name == "baseline":
                if acquisition_type == AcquisitionType.SNAPSHOT_AT_START:
                    # Instead of directly indexing seq_num, get the lower seq_num value may change
                    # depending in the order of acquisition
                    first_seq_num = min(retrieved_stream_data.keys())
                    collapsed_retrieved_stream_data = self._merge_by_seq_num(
                        {first_seq_num: retrieved_stream_data[first_seq_num]}
                    )
                elif acquisition_type == AcquisitionType.SNAPSHOT_AT_END:
                    # Instead of directly indexing seq_num, get the higher seq_num returned as seq_num value may change
                    # depending in the order of acquisition
                    last_seq_num = max(retrieved_stream_data.keys())
                    collapsed_retrieved_stream_data = self._merge_by_seq_num(
                        {last_seq_num: retrieved_stream_data[last_seq_num]}
                    )
                else:
                    raise AssonantDataRetrieverError(
                        f"Invalid AcquisitionType '{acquisition_type.value}' for 'baseline' Event Stream"
                    )
            else:
                collapsed_retrieved_stream_data = self._merge_by_seq_num(retrieved_stream_data)

            data_retrived_by_query = data_retrived_by_query | collapsed_retrieved_stream_data

        # Convert data representation to fit the expected return structure
        formatted_retrieved_data = {}

        for pv_name, pv_info_dict in data_retrived_by_query.items():
            formatted_retrieved_data[pv_name] = pv_info_dict["data"]["value"]

        return formatted_retrieved_data



"""Assonant Path Builder.

Assonant Path builder submodule defines Builders classes responsible to
centralize and standardize path mounting tasks inside Assonant. With this
submodule, different classes to handle specific directories structures
within or outside Sirius can be handled by those classes and made easy to
maintain. 

The idea is to follow the Builder Design pattern within the classes in this
submodule.

Reference: https://refactoring.guru/design-patterns/builder
"""

from .beamline_path_builder import BeamlinePathBuilder

__all__ = ["BeamlinePathBuilder"]



from abc import ABC
from pathlib import Path
from typing import Callable, Dict, List

from ._csv_metadata_retriever import CSVMetadataRetriever
from .assonant_metadata_retriever_interface import IAssonantMetadataRetriever
from .enums import MetadataSourceFileFormat
from .exceptions import MetadataRetrieverFactoryError


def get_file_format(data_source_file_path: str) -> str:
    """Extract file extension from the given path.

    Args:
        data_source_file_path (str): Path to data source file.

    Raises:
        FileNotFoundError: If the path is invalid or not a file.

    Returns:
        str: File extension without '.' character.
    """
    path = Path(data_source_file_path)
    if not path.is_file():
        raise FileNotFoundError(f"'{data_source_file_path}' is not a valid file!")
    return path.suffix[1:]


class MetadataRetrieverFactory(ABC):
    """Abstract Factory for creating Assonant Metadata Retrievers.

    Class that implements the factory design pattern
    (https://refactoring.guru/design-patterns/factory-method) to fully abstract
    the procedure of creating Assonant Metadata Retrievers.

    Uses a registry-based approach so that new metadata retrievers can be added
    without modifying the core creation logic (Open/Closed principle).

    Example:
        MetadataRetrieverFactory.register_file_format("csv", CSVMetadataRetriever)
        retriever = MetadataRetrieverFactory.create_metadata_retriever("data.csv")
    """

    _creators: Dict[str, Callable[[str], IAssonantMetadataRetriever]] = {}

    @staticmethod
    def register_file_format(file_format: str, creator: Callable[[str], IAssonantMetadataRetriever]) -> None:
        """Register a file format with its creator callable.

        Args:
            file_format (str): File format key (lowercase recommended).
            creator (Callable[[str], IAssonantMetadataRetriever]): Function or class
                returning an IAssonantMetadataRetriever instance.

        Raises:
            MetadataRetrieverFactoryError: If passed creator is not a callable object.

        """
        if not callable(creator):
            raise MetadataRetrieverFactoryError(
                "Creator must be callable and return an IAssonantMetadataRetriever instance."
            )

        MetadataRetrieverFactory._creators[file_format] = creator

    @staticmethod
    def create_metadata_retriever(metadata_source_file_path: str) -> IAssonantMetadataRetriever:
        """Create a metadata retriever instance for the given file.

        Args:
            metadata_source_file_path (str): Path to metadata source file.

        Raises:
            MetadataRetrieverFactoryError: If the file format is unsupported or invalid.

        Returns:
            IAssonantMetadataRetriever: Metadata retriever instance.
        """
        try:
            file_format = get_file_format(metadata_source_file_path)
        except Exception:
            raise MetadataRetrieverFactoryError(f"'{metadata_source_file_path}' is not a file!")

        try:
            creator = MetadataRetrieverFactory._creators[file_format]
        except KeyError:
            raise MetadataRetrieverFactoryError(
                f"'{file_format}' file format isn't currently supported. "
                f"Supported formats: {list(MetadataRetrieverFactory._creators.keys())}"
            )

        return creator(metadata_source_file_path)

    @staticmethod
    def get_supported_file_formats() -> List[str]:
        """Return a list of supported file formats."""
        return list(MetadataRetrieverFactory._creators.keys())


# === Register built-in formats here ===
MetadataRetrieverFactory.register_file_format(MetadataSourceFileFormat.CSV.value, CSVMetadataRetriever)



from typing import Any, Dict, List, Union

from assonant.naming_standards import AcquisitionType

from .assonant_data_retriever_interface import IAssonantDataRetriever
from .data_retriever_factory import DataRetrieverFactory
from .exceptions.data_retrievers_exceptions import AssonantDataRetrieverError


class AssonantDataRetriever(IAssonantDataRetriever):
    """Assonant Data Retriever.

    Wrapper class that abstracts all process related to creating specific data retrievers
    for specific file formats/ data structures and instatiating components and fields given the retrieved data.
    """

    def __init__(self, data_source: Any):

        # Persist passed data_source
        self.data_source = data_source
        try:
            self.data_retriever = DataRetrieverFactory.create_data_retriever(data_source=data_source)
        except Exception as e:
            raise AssonantDataRetrieverError(
                f"Failed to create a suitable DataRetriever for data_source of type: {type(data_source)}"
            ) from e

    def get_pv_data_by_acquisition_type(self, acquisition_type: AcquisitionType) -> Dict[str, Union[Any, List[Any]]]:
        """Return collected PV data related to specified AcquisitionType.

        The returned value is a dictionary as follow:

        {
            PV_NAME_1: [PV_VALUE_1, PV_VALUE_2, ...],
            PV_NAME_2: [PV_VALUE_1],
            ...
        }

        PS: PV values are always returned within a list, even if it is a single value.

        Args:
            acquisition_type (AcquisitionType): Target Acquisition Type used to select
            data from which Acquisition Type will be retrieved.

        Returns:
            Dict[str, Union[Any, List[Any]]]: Dictionary containing retrieved PV data.
        """
        try:
            return self.data_retriever.get_pv_data_by_acquisition_type(acquisition_type)
        except Exception as e:
            raise AssonantDataRetrieverError(
                f"Failed to retrieve pv data for '{acquisition_type.value}' Acquisition Type"
            ) from e



from abc import ABC
from pathlib import Path
from typing import Any, Callable, Dict, List

from ._bluesky_data_retriever import BlueskyDataRetriever
from .assonant_data_retriever_interface import IAssonantDataRetriever
from .enums import DataSourceFileFormat
from .exceptions import AssonantDataRetrieverError


def get_file_format(data_source_file_path: str) -> str:
    """Extract file extension from the given path.

    Args:
        data_source_file_path (str): Path to data source file.

    Raises:
        AssonantDataRetrieverError: If the path is invalid or not a file.

    Returns:
        str: File extension without '.' character.
    """
    path = Path(data_source_file_path)
    if not path.is_file():
        raise AssonantDataRetrieverError(f"'{data_source_file_path}' is not a valid file!")
    return path.suffix[1:]


class DataRetrieverFactory(ABC):
    """Abstract Factory for creating Assonant Data Retrievers.

    Class that implements the factory design pattern
    (https://refactoring.guru/design-patterns/factory-method) to fully abstract
    the procedure of creating Assonant Metadata Retrievers.

    Uses a registry-based approach so that new data retrievers can be added
    without modifying the core creation logic (Open/Closed principle).

    Example:
        DataRetrieverFactory.register_file_format("json", BlueskyDataRetriever)
        retriever = DataRetrieverFactory.create_data_retriever("data.json")
    """

    _creators: Dict[str, Callable[[Any], IAssonantDataRetriever]] = {}

    @staticmethod
    def register_file_format(file_format: str, creator: Callable[[Any], IAssonantDataRetriever]) -> None:
        """Register a file format with its creator callable.

        Args:
            file_format (str): File format key (lowercase recommended).
            creator (Callable[[Any], IAssonantDataRetriever]): Function or class
                returning an IAssonantDataRetriever instance.

        Raises:
            AssonantDataRetrieverError: If passed creator is not callable.
        """
        if not callable(creator):
            raise AssonantDataRetrieverError("Creator must be callable and return an IAssonantDataRetriever instance.")

        DataRetrieverFactory._creators[file_format] = creator

    @staticmethod
    def create_data_retriever(data_source: Any) -> IAssonantDataRetriever:
        """Create a data retriever instance for the given data source.

        Args:
            data_source (Any): Data source to retrieve data from.
                - str: File path (extension determines retriever type).
                - list: Already loaded Bluesky documents.

        Raises:
            AssonantDataRetrieverError: If the file format or data source type is unsupported.

        Returns:
            IAssonantDataRetriever: Data retriever instance.
        """
        if isinstance(data_source, str):
            # File path case
            try:
                file_format = get_file_format(data_source)
            except Exception:
                raise AssonantDataRetrieverError(f"'{data_source}' is not a valid file!")

            try:
                creator = DataRetrieverFactory._creators[file_format]
            except KeyError:
                raise AssonantDataRetrieverError(
                    f"'{file_format}' file format isn't currently supported. "
                    f"Supported formats: {list(DataRetrieverFactory._creators.keys())}"
                )

            return creator(data_source)

        elif isinstance(data_source, list):
            # In-memory Bluesky data
            try:
                creator = DataRetrieverFactory._creators[DataSourceFileFormat.JSON.value]
            except KeyError:
                raise AssonantDataRetrieverError(
                    f"Bluesky in-memory data requires JSON retriever, "
                    f"but it is not registered! Supported formats: "
                    f"{list(DataRetrieverFactory._creators.keys())}"
                )
            return creator(data_source)

        else:
            raise AssonantDataRetrieverError("Data source type not supported for data retrieving!")

    @staticmethod
    def get_supported_file_formats() -> List[str]:
        """Return a list of supported file formats."""
        return list(DataRetrieverFactory._creators.keys())


# === Register built-in retrievers here ===
DataRetrieverFactory.register_file_format(DataSourceFileFormat.JSON.value, BlueskyDataRetriever)



"""Exceptions triggered to warn about internal errors related to Assonant data retriever classes."""


class AssonantDataRetrieverError(Exception):
    """Custom Exception to warn when any error occurs inside AssonantDataRetriever."""



"""Assonant metadata retriever enums.

This submodule defines Enumerations classes used to standardize options related to metadata retrieving.
"""

from .csv_column import CSVColumn
from .metadata_source_file_format import MetadataSourceFileFormat

__all__ = ["CSVColumn", "MetadataSourceFileFormat"]



from enum import Enum


class ExperimentStage(Enum):
    """Possible Experiment Stage values."""

    SAMPLE_ACQUISITION = "sample_acquisition"
    SAMPLE_POSITIONING = "sample_positioning"
    FLAT_FIELD_ACQUISITION = "flat_field_acquisition"
    DARK_FIELD_ACQUISITION = "dark_field_acquisition"
    BUFFER_ACQUISITION = "buffer_acquisition"



from typing import List

from ._assonant_data_sender_interface import IAssonantDataSender
from ._kafka_data_sender import KafkaDataSender
from .exceptions import AssonantDataSenderError


class DataSenderFactory:
    """Data Sender Factory.

    Class that implements the factory design pattern
    (https://refactoring.guru/design-patterns/factory-method) to fully abstract
    the procedure of creating Assonant Data senders
    """

    _supported_comm_methods = ["kafka"]

    def create_data_sender(self, comm_method: str, configs: dict) -> IAssonantDataSender:
        """Public method that abstracts data sender creation process for the factory user.

        Internally, this method deals with validation and specific Data Sender
        creation

        Args:
            comm_method (str): Communication method used by the Data Sender that
            will be created.
            configs (dict): Dictionary containing all configuration settings
            from the specific Data Sender that will created. Different
            communication methods may require different settings and fields.

        Raises:
            AssonantDataSenderError: An error occured during the creation of
            the respective Data Sender.

        Returns:
            IAssonantDataSender: Data Sender instance which implements the
            IAssonantDataSender interface
        """
        self._validate_comm_method(comm_method)

        if comm_method == "kafka":
            return self._create_kafka_data_sender(configs)
        raise AssonantDataSenderError(
            f"'{comm_method}' communication method is set as supported but its creation method is not implemented."
        )

    def _create_kafka_data_sender(self, configs: dict) -> KafkaDataSender:
        return KafkaDataSender(configs)

    def _validate_comm_method(self, comm_method: str) -> None:
        """Validate if passed communication method is currently supported.

        Args:
            comm_method (str): Communication method to be validated

        Raises:
            AssonantDataSenderError: Communication method not supported.
        """
        if comm_method not in self._supported_comm_methods:
            raise AssonantDataSenderError(
                f"'{comm_method}' communication method isn't currently supported. "
                f"The supported methods are: {self._supported_comm_methods}"
            )

    def get_supported_comm_methods(self) -> List[str]:
        """Getter method for current supported communication methods.

        Returns:
            List[str]: List containing curretly supported communication methods.
        """
        return self._supported_comm_methods



from pathlib import Path


class BeamlinePathBuilder:
    """A utility class to build standardized filesystem paths for beamline-related data and proposals.

    Attributes:
        _root (Path): The root directory for all beamline paths.
        _data_schema_file_name (str): The name of the data schema metadata file.
        beamline_name (str): The name of the beamline (must be set before building paths).
        proposal_id (str): The ID of the proposal (optional, but required for some paths).
    """

    def __init__(self):
        """Initializes the BeamlinePathBuilder with default root and data schema filename."""
        self._root = Path("/ibira/lnls/beamlines")
        self._data_schema_file_name = "metadata.csv"
        self.beamline_name = None
        self.proposal_id = None

    def set_beamline_name(self, beamline_name: str):
        """Sets the beamline name for which paths will be built.

        Args:
            beamline_name (str): The name of the beamline.
        """
        self.beamline_name = beamline_name

    def set_proposal(self, proposal_id: str):
        """Sets the proposal ID for which paths will be built.

        Args:
            proposal_id (str): The ID of the proposal.
        """
        self.proposal_id = proposal_id

    def build_path_to_data_schema(self):
        """Builds the path to the data schema metadata file for the specified beamline.

        Returns:
            Path: The full path to the data schema file.

        Raises:
            AssertionError: If beamline_name is not set.
        """
        assert (
            self.beamline_name is not None
        ), "Beamline name property must be set to allow building path to data schema file!"

        path_to_data_schema = self._root / self.beamline_name / Path("apps/assonant") / self._data_schema_file_name
        return str(path_to_data_schema)

    def build_path_to_proposal(self):
        """Builds the path to the directory for the proposal ID previously set.

        Returns:
            Path: The full path to the specific proposal ID directory.

        Raises:
            AssertionError: If beamline_name or proposal_id is not set.
        """
        assert (
            self.beamline_name is not None
        ), "Beamline name property must be set to allow building path to proposal directory!"
        assert (
            self.proposal_id is not None
        ), "Proposal ID property must be set to allow building path to proposal directory!"

        path_to_proposal = self._root / self.beamline_name / Path("proposals") / self.proposal_id
        return str(path_to_proposal)

    def reset(self):
        """Resets the beamline name and proposal ID to None.
        Useful for reusing the builder for different configurations.
        """
        self.beamline_name = None
        self.proposal_id = None
